﻿
namespace NiknakV2.Business.Gallery.Drawing
{
    public class ImageStack
    {

        #region Fields

        public string FolderImagePath { get; set; }
        public int MaxDimension { get; set; }
        private string PhysicalPath { get; set; }
        private bool _InError;

        #endregion

        #region Properties

        public bool InError
        {
            get { return _InError; }
        }

        public Picture Picture1 { get; set; }
        public Picture Picture2 { get; set; }
        public Picture Picture3 { get; set; }

        #endregion

        #region Methods

        public ImageStack(string physicalPath)
        {
            PhysicalPath = physicalPath;
        }

        private static byte[] LoadImageData(string filePath)
        {
            if (System.IO.File.Exists(filePath))
            {
                FileStream fs = new FileStream(filePath, FileMode.Open, FileAccess.Read);
                BinaryReader br = new BinaryReader(fs);
                byte[] imageBytes = br.ReadBytes((int)fs.Length);
                br.Close();
                fs.Close();
                return imageBytes;
            }
            
            return new byte[0];
        }

        private static BitmapSource CreateImage(byte[] imageData, int decodePixelWidth, int decodePixelHeight)
        {
            if (imageData == null) return null;

            BitmapImage result = new BitmapImage();
            result.BeginInit();
            if (decodePixelWidth > 0)
            {
                result.DecodePixelWidth = decodePixelWidth;
            }
            if (decodePixelHeight > 0)
            {
                result.DecodePixelHeight = decodePixelHeight;
            }
            result.StreamSource = new MemoryStream(imageData);
            result.CreateOptions = BitmapCreateOptions.None;
            result.CacheOption = BitmapCacheOption.Default;
            result.EndInit();
            return result;
        }

        public Canvas RenderSingleImage(Picture sourcePicture, double skewAngle, double shadowHeight, double shadowLightAngle)
        {
            //setup the drawing area
            Canvas myCanvas = new Canvas();
            myCanvas.Background = Brushes.Transparent;

            //load up the specified image
            System.Windows.Controls.Image baseImage = new System.Windows.Controls.Image();
            byte[] imageBytes;
            int loadWidth = MaxDimension;

            if (sourcePicture != null)
                imageBytes = LoadImageData(sourcePicture.FullName);
            else
            {
                loadWidth = 0;
                imageBytes = LoadImageData(FolderImagePath);
            }

            BitmapSource imageSource = null;
            if (!imageBytes.Length.Equals(0))
                imageSource = CreateImage(imageBytes, loadWidth, 0);
            imageBytes = null;

            //untested - ignore landscape images
            if (imageSource != null && imageSource.PixelWidth < imageSource.PixelHeight)
            {
                imageSource = null;
                sourcePicture = null;
            }

            double reductionFactor = ((double)1 / (double)3) * (double)2;
            double newHeight;
            if (sourcePicture != null)
            {
                //force the image into 1.5:1 aspect ratio (as used by SLR cameras)
                //nb: some cameras seem to use 1.48:1 and some use 1.51:1, which doesn't really help
                newHeight = (double)imageSource.PixelWidth * reductionFactor;
                //in case the ratio is less than 1.5:1 reset height
                if (newHeight > (double)imageSource.PixelHeight)
                    newHeight = imageSource.PixelHeight;
                double cropHeight = ((double)imageSource.PixelHeight - newHeight) / (double)2;
                CroppedBitmap croppedSource = new CroppedBitmap(imageSource,
                    new Int32Rect(
                        0, //origin x
                        (int)Math.Round(cropHeight, 0, MidpointRounding.AwayFromZero), //origin y
                        imageSource.PixelWidth, //width
                        (int)Math.Round(newHeight, 0, MidpointRounding.AwayFromZero) //height
                    )
                );

                baseImage.Source = croppedSource;
            }
            else
                baseImage.Source = imageSource;

            //configure image control
            newHeight = (double)MaxDimension * reductionFactor;
            baseImage.Width = MaxDimension;
            baseImage.Height = newHeight;

            //create a border to house the image
            Border baseBorder = new Border();
            baseBorder.Background = Brushes.White; //ensures any space is opaque
            baseBorder.Child = baseImage;
            baseBorder.BorderBrush = Brushes.Gray;
            baseBorder.BorderThickness = new Thickness(((double)MaxDimension / (double)100) * (double)1.5);
            baseBorder.HorizontalAlignment = HorizontalAlignment.Left;
            baseBorder.VerticalAlignment = VerticalAlignment.Top;

            //apply skew transform to border/image
            baseBorder.RenderTransform = new SkewTransform(0, skewAngle);
            
            //create a border control to use as a shadow
            Border shadowBorder = new Border();
            shadowBorder.Width = baseImage.Width + (baseBorder.BorderThickness.Top * 2);
            shadowBorder.Height = shadowHeight;
            shadowBorder.Background = Brushes.Black;

            //build and apply a gradient to the border
            LinearGradientBrush linGradient = new LinearGradientBrush();
            linGradient.StartPoint = new Point(0, 0);
            linGradient.EndPoint = new Point(0, 1);
            linGradient.GradientStops.Add(new GradientStop(Colors.Black, 0));
            linGradient.GradientStops.Add(new GradientStop(Colors.Transparent, 0.47));
            shadowBorder.OpacityMask = linGradient;

            //apply transformations to shadow
            TransformGroup shadowTranGroup = new TransformGroup();
            shadowTranGroup.Children.Add(new RotateTransform(180, shadowBorder.Width/2, 0));
            shadowTranGroup.Children.Add(new SkewTransform(shadowLightAngle, skewAngle));
            shadowBorder.RenderTransform = shadowTranGroup;

            //use a blur to make the shadow softer
            BlurEffect shadowBlur = new BlurEffect();
            shadowBlur.Radius = 15;
            shadowBorder.Effect = shadowBlur;

            //position the shadow so it appears to be cast by the main image
            shadowBorder.SetValue(Canvas.LeftProperty, (double)0);
            shadowBorder.SetValue(Canvas.TopProperty, (double)(baseImage.Height/* + baseBorder.BorderThickness.Top*/));

            //add shadow and image to canvas
            myCanvas.Children.Add(shadowBorder);
            myCanvas.Children.Add(baseBorder);

            //calculate the additional height created by the skew of the main image
            double canvasHeight = baseImage.Height + (baseBorder.BorderThickness.Top * (double)2);
            double sideB = baseImage.Width + (baseBorder.BorderThickness.Top * (double)2);
            double angleA = (((double)2 * Math.PI) / (double)360 * skewAngle);
            double skewHeightAddition = Math.Tan(angleA) * sideB;

            //calculate the additional width created by the skew of the shadow
            double canvasWidth = baseImage.Width + (baseBorder.BorderThickness.Top * 2); 
            double sideC = shadowHeight;
            angleA = ((2 * Math.PI) / 360 * (90 - (360 - shadowLightAngle)));
            double skewWidthAddition = Math.Cos(angleA) * sideC; 

            //trim the shadow width addition since the shadow only uses about 70% of the shadowWidth
            skewWidthAddition = (skewWidthAddition / (double)100) * (double)70;

            //explicitly set the canvas dimensions since it can't be trusted to calculate its own
            myCanvas.Width = Math.Round(canvasWidth + skewWidthAddition, MidpointRounding.AwayFromZero);
            myCanvas.Height = Math.Round(canvasHeight + skewHeightAddition, MidpointRounding.AwayFromZero);

            //rendering stuff
            myCanvas.Measure(new Size(myCanvas.Width, myCanvas.Height));
            myCanvas.Arrange(new Rect(new Size(myCanvas.Width, myCanvas.Height)));

            return myCanvas;
        }

        public byte[] Render()
        {
            //declarations used in try and finally
            byte[] output;
            JpegBitmapEncoder jpgEncoder = null;
            RenderTargetBitmap renderedCanvas = null;
            Canvas pic1 = null, pic2 = null, pic3 = null;
            Canvas rootCanvas = null;

            try
            {
                //first check for a cached ImageStack
                if (File.Exists(CacheFileName))
                {
                    //cache file already exists so don't bother making a new image
                    FileStream cachedStack = new FileStream(CacheFileName, FileMode.Open, FileAccess.Read);
                    output = new byte[cachedStack.Length];
                    cachedStack.Read(output, 0, (int)cachedStack.Length);
                    cachedStack.Dispose();
                }
                else
                {
                    Album myAlbum = new Album(PhysicalPath);

                    int skewMax = 7;
                    int skewMed = 5;
                    int skewMin = 3;

                    rootCanvas = new Canvas();
                    rootCanvas.Background = Brushes.White;

                    double shadowHeight = ((double)MaxDimension / (double)100) * (double)30; //shadowHeight is 30% of MaxDimension
                    double margin = ((double)MaxDimension / (double)100) * (double)17; //margin is 17% of the width of each image before transforms

                    //load first pic
                    pic1 = RenderSingleImage(Picture1, skewMax, shadowHeight, 320);

                    //load additional pics if available
                    if (Picture2 != null)
                        pic2 = RenderSingleImage(Picture2, skewMed, shadowHeight, 320);
                    if (Picture3 != null)
                        pic3 = RenderSingleImage(Picture3, skewMin, shadowHeight, 320);

                    //position pics into stack effect, using bottom right allows for future support for portrait images
                    if (pic2 != null && pic3 != null)
                    {
                        Canvas.SetRight(pic1, (double)(margin * 2));
                        Canvas.SetBottom(pic1, (double)0);
                        Canvas.SetRight(pic2, (double)margin);
                        Canvas.SetBottom(pic2, (double)margin);
                        Canvas.SetRight(pic3, (double)0);
                        Canvas.SetBottom(pic3, (double)(margin * 2));

                        rootCanvas.Children.Add(pic3);
                        rootCanvas.Children.Add(pic2);
                        rootCanvas.Children.Add(pic1);
                    }
                    //just two pics
                    else if (pic2 != null)
                    {
                        Canvas.SetRight(pic1, (double)(margin * 1.5));
                        Canvas.SetBottom(pic1, (double)(margin / 3));
                        Canvas.SetRight(pic2, (double)(margin / 2));
                        Canvas.SetBottom(pic2, (double)(margin + (margin / 3)));

                        rootCanvas.Children.Add(pic2);
                        rootCanvas.Children.Add(pic1);

                        //for purposes of canvas dimensions, render the third (least skewed) image
                        pic3 = RenderSingleImage(null, skewMin, shadowHeight, 320);
                    }
                    //just one pic
                    else
                    {
                        Canvas.SetRight(pic1, ((double)(margin)));
                        Canvas.SetBottom(pic1, (double)((margin / 4) * 3));

                        rootCanvas.Children.Add(pic1);

                        //for purposes of canvas dimensions, render the third (least skewed) image
                        pic3 = RenderSingleImage(null, skewMin, shadowHeight, 320);
                    }
                    rootCanvas.UpdateLayout();

                    int Width = (int)(pic1.ActualWidth + Math.Round((double)margin * 2, MidpointRounding.AwayFromZero));
                    int Height = (int)(pic3.ActualHeight + Math.Round((double)margin * 2, MidpointRounding.AwayFromZero));

                    rootCanvas.Width = Width;
                    rootCanvas.Height = Height;

                    //rendering stuff
                    rootCanvas.Measure(new Size((int)rootCanvas.Width, (int)rootCanvas.Height));
                    rootCanvas.Arrange(new Rect(new Size((int)rootCanvas.Width, (int)rootCanvas.Height)));

                    //first render the canvas into a bitmap so we can scale it
                    renderedCanvas = new RenderTargetBitmap(Width, Height, 96, 96, PixelFormats.Default);
                    renderedCanvas.Render(rootCanvas);

                    jpgEncoder = new JpegBitmapEncoder();
                    jpgEncoder.Frames.Add(BitmapFrame.Create(renderedCanvas));

                    using (MemoryStream ms = new MemoryStream())
                    {
                        jpgEncoder.Save(ms);
                        output = ms.ToArray();
                    }

#if !DEBUG
                    //cache the output to disk
                    File.WriteAllBytes(CacheFileName, output);
#endif
                }

                _InError = false;
            }
            catch (Exception ex)
            {
                output = CatchError(ex.Message);
                _InError = true;
            }
            finally
            {
                //be very specific about clean up
                if (jpgEncoder != null && jpgEncoder.Dispatcher.Thread.IsAlive)
                    jpgEncoder.Dispatcher.InvokeShutdown();
                if (renderedCanvas != null && renderedCanvas.Dispatcher.Thread.IsAlive)
                    renderedCanvas.Dispatcher.InvokeShutdown();

                jpgEncoder = null;
                renderedCanvas = null;

                pic1 = null;
                pic2 = null;
                pic3 = null;
                rootCanvas = null;
            }

            return output;
        }

        private byte[] CatchError(string message)
        {
            _InError = true;
            return ImageError.RenderAsByteArray(message);
        }

        #endregion
    }
}
